@@ -27,57 +27,84 @@ from typing import Dict
 from typing import List
 from typing import Optional
 from typing import Tuple
-from typing import Union
 
 
-__version__ = '0.0.4'
+__version__ = '0.0.7'
 
 
 DEFAULT_BRANCH = 'master'
 DEFAULT_REMOTE = 'origin'
+ENV_VAR_CHECK_MERGE_COMMITS = 'DCO_CHECK_CHECK_MERGE_COMMITS'
+ENV_VAR_DEFAULT_BRANCH = 'DCO_CHECK_DEFAULT_BRANCH'
+ENV_VAR_DEFAULT_BRANCH_FROM_REMOTE = 'DCO_CHECK_DEFAULT_BRANCH_FROM_REMOTE'
+ENV_VAR_DEFAULT_REMOTE = 'DCO_CHECK_DEFAULT_REMOTE'
+ENV_VAR_QUIET = 'DCO_CHECK_QUIET'
+ENV_VAR_VERBOSE = 'DCO_CHECK_VERBOSE'
 TRAILER_KEY_SIGNED_OFF_BY = 'Signed-off-by:'
 
 
-class BooleanDefaultValue:
-    """Default value wrapper for bool."""
+class EnvDefaultOption(argparse.Action):
+    """
+    Action that uses an env var value as the default if it exists.
 
-    def __init__(self, value):  # noqa: D107
-        assert(isinstance(value, bool))
-        self._value = value
+    Inspired by: https://stackoverflow.com/a/10551190/6476709
+    """
 
-    def __bool__(self):  # noqa: D105
-        return self._value
+    def __init__(
+        self,
+        env_var: str,
+        default: Any,
+        help=None,  # noqa: A002
+        **kwargs,
+    ) -> None:
+        """Create a EnvDefaultOption."""
+        # Set default to env var value if it exists
+        if env_var in os.environ:
+            default = os.environ[env_var]
+        if help:  # pragma: no cover
+            help += f' [env var: {env_var}]'
+        super(EnvDefaultOption, self).__init__(
+            default=default,
+            help=help,
+            **kwargs,
+        )
+
+    def __call__(self, parser, namespace, values, option_string=None):  # noqa
+        setattr(namespace, self.dest, values)
 
 
-class StringDefaultValue(str):
-    """Default value wrapper for str."""
+class EnvDefaultStoreTrue(argparse.Action):
+    """
+    Action similar to 'store_true' that uses an env var value as the default if it exists.
 
-    pass
+    Partly copied from arparse.{_StoreConstAction,_StoreTrueAction}.
+    """
 
+    def __init__(
+        self,
+        option_strings,
+        dest,
+        env_var: str,
+        default: bool = False,
+        help=None,  # noqa: A002
+    ) -> None:
+        """Create a EnvDefaultStoreTrue."""
+        # Set default value to true if the env var exists
+        default = env_var in os.environ
+        if help:  # pragma: no cover
+            help += f' [env var: {env_var} (set to anything)]'
+        super(EnvDefaultStoreTrue, self).__init__(
+            option_strings=option_strings,
+            dest=dest,
+            nargs=0,
+            const=True,
+            default=default,
+            required=False,
+            help=help,
+        )
 
-_types = {
-    bool: BooleanDefaultValue,
-    str: StringDefaultValue,
-}
-
-
-def wrap_default_value(
-    value: Union[bool, str],
-) -> Union[BooleanDefaultValue, StringDefaultValue, Any]:
-    """Wrap a default value so that we can check if a value is still the default."""
-    # Inspired by: https://github.com/colcon/colcon-core/pull/288/files
-    global _types
-    if is_default_value(value):
-        raise ValueError('tried to wrap a default value a second time')
-    if type(value) in _types:
-        return _types[type(value)](value)
-    return value
-
-
-def is_default_value(value: Any) -> bool:
-    """Check if a value is a default value."""
-    global _types
-    return isinstance(value, tuple(_types.values()))
+    def __call__(self, parser, namespace, values, option_string=None):  # noqa
+        setattr(namespace, self.dest, self.const)
 
 
 def get_parser() -> argparse.ArgumentParser:
@@ -85,41 +112,68 @@ def get_parser() -> argparse.ArgumentParser:
     parser = argparse.ArgumentParser(
         description='Check that all commits of a proposed change have a DCO, i.e. are signed-off.',
     )
-    parser.add_argument(
-        '-b', '--default-branch',
-        default=wrap_default_value(DEFAULT_BRANCH),
-        help='default branch to use, if necessary (default: %(default)s)',
+    default_branch_group = parser.add_mutually_exclusive_group()
+    default_branch_group.add_argument(
+        '-b', '--default-branch', metavar='BRANCH',
+        action=EnvDefaultOption, env_var=ENV_VAR_DEFAULT_BRANCH,
+        default=DEFAULT_BRANCH,
+        help=(
+            'default branch to use, if necessary (default: %(default)s)'
+        ),
+    )
+    default_branch_group.add_argument(
+        '--default-branch-from-remote',
+        action=EnvDefaultStoreTrue, env_var=ENV_VAR_DEFAULT_BRANCH_FROM_REMOTE,
+        default=False,
+        help=(
+            'get the default branch value from the remote (default: %(default)s)'
+        ),
     )
     parser.add_argument(
         '-m', '--check-merge-commits',
-        action='store_true',
-        default=wrap_default_value(False),
-        help='check sign-offs on merge commits as well',
+        action=EnvDefaultStoreTrue, env_var=ENV_VAR_CHECK_MERGE_COMMITS,
+        default=False,
+        help=(
+            'check sign-offs on merge commits as well (default: %(default)s)'
+        ),
     )
     parser.add_argument(
-        '-r', '--default-remote',
-        default=wrap_default_value(DEFAULT_REMOTE),
-        help='default remote to use, if necessary (default: %(default)s)',
+        '-r', '--default-remote', metavar='REMOTE',
+        action=EnvDefaultOption, env_var=ENV_VAR_DEFAULT_REMOTE,
+        default=DEFAULT_REMOTE,
+        help=(
+            'default remote to use, if necessary (default: %(default)s)'
+        ),
     )
     output_options_group = parser.add_mutually_exclusive_group()
     output_options_group.add_argument(
         '-q', '--quiet',
-        action='store_true',
-        default=wrap_default_value(False),
-        help='quiet mode (do not print anything; simply exit with 0 or non-zero)',
+        action=EnvDefaultStoreTrue, env_var=ENV_VAR_QUIET,
+        default=False,
+        help=(
+            'quiet mode (do not print anything; simply exit with 0 or non-zero) '
+            '(default: %(default)s)'
+        ),
     )
     output_options_group.add_argument(
         '-v', '--verbose',
-        action='store_true',
-        default=wrap_default_value(False),
-        help='verbose mode (print out more information)',
+        action=EnvDefaultStoreTrue, env_var=ENV_VAR_VERBOSE,
+        default=False,
+        help=(
+            'verbose mode (print out more information) (default: %(default)s)'
+        ),
     )
     return parser
 
 
-def parse_args() -> argparse.Namespace:
-    """Parse arguments."""
-    return get_parser().parse_args()
+def parse_args(argv: Optional[List[str]] = None) -> argparse.Namespace:
+    """
+    Parse arguments.
+
+    :param argv: the arguments to use, or `None` for sys.argv
+    :return: the parsed arguments
+    """
+    return get_parser().parse_args(argv)
 
 
 class Options:
@@ -129,6 +183,7 @@ class Options:
         """Create using default argument values."""
         self.check_merge_commits = parser.get_default('m')
         self.default_branch = parser.get_default('b')
+        self.default_branch_from_remote = parser.get_default('default-branch-from-remote')
         self.default_remote = parser.get_default('r')
         self.quiet = parser.get_default('q')
         self.verbose = parser.get_default('v')
@@ -137,62 +192,29 @@ class Options:
         """Set options using parsed arguments."""
         self.check_merge_commits = args.check_merge_commits
         self.default_branch = args.default_branch
+        self.default_branch_from_remote = args.default_branch_from_remote
         self.default_remote = args.default_remote
         self.quiet = args.quiet
         self.verbose = args.verbose
-
-    def apply_env_vars(self) -> None:
-        """
-        Apply environment variable values.
-
-        Only uses an environment variable value if the argument has not been set through CLI.
-        """
-        def real_cast(value: Any, param_type: Any) -> Any:
-            """Because bool('False') is True."""
-            if param_type is bool:
-                return value in [1, '1', 'true', 'True', 'y', 'Y', 'yes', 'Yes']
-            return param_type(value)
-
-        def environment_value_if_default_param(
-            param_name: str,
-            param_type: Any,
-            env_var_name: str,
-        ) -> Optional[Any]:
-            """
-            Choose value between environment variable and CLI argument.
-
-            Returns environment variable value if it exists and the parameter has a default value.
-            """
-            if not hasattr(self, param_name):
-                raise ValueError('parameter name does not exist')
-            value = getattr(self, param_name)
-            # Use environment variable value if it exists, otherwise keep default value
-            if is_default_value(value):
-                return real_cast(
-                    get_env_var(env_var_name, print_if_not_found=False, default=value),
-                    param_type,
-                )
-            else:
-                return real_cast(value, param_type)
-        self.check_merge_commits = environment_value_if_default_param(
-            'check_merge_commits',
-            bool,
-            'DCO_CHECK_CHECK_MERGE_COMMITS',
-        )
-        self.default_branch = environment_value_if_default_param(
-            'default_branch',
-            str,
-            'DCO_CHECK_DEFAULT_BRANCH',
-        )
-        self.default_remote = environment_value_if_default_param(
-            'default_remote',
-            str,
-            'DCO_CHECK_DEFAULT_REMOTE',
-        )
-        self.quiet = environment_value_if_default_param('quiet', bool, 'DCO_CHECK_QUIET')
-        self.verbose = environment_value_if_default_param('verbose', bool, 'DCO_CHECK_VERBOSE')
+        # Shouldn't happen with a mutually exclusive group,
+        # but can happen if one is set with an env var
+        # and the other is set with an arg
         if self.quiet and self.verbose:
-            raise ValueError("'quiet' and 'verbose' cannot both be true")
+            # Similar message to what is printed when using args for both
+            get_parser().print_usage()
+            print("options '--quiet' and '--verbose' cannot both be true")
+            sys.exit(1)
+        if self.default_branch != DEFAULT_BRANCH and self.default_branch_from_remote:
+            # Similar message to what is printed when using args for both
+            get_parser().print_usage()
+            print(
+                "options '--default-branch' and '--default-branch-from-remote' cannot both be set"
+            )
+            sys.exit(1)
+
+    def get_options(self) -> Dict:
+        """Get all options as a dict."""
+        return self.__dict__
 
 
 options = Options(get_parser())
@@ -324,6 +346,32 @@ def fetch_branch(
     return 0 if run(command) is not None else 1
 
 
+def get_default_branch_from_remote(
+    remote: str,
+) -> Optional[str]:
+    """
+    Get default branch from remote.
+
+    :param remote: the remote name
+    :return: the default branch, or None if it failed
+    """
+    # https://stackoverflow.com/questions/28666357/git-how-to-get-default-branch#comment92366240_50056710  # noqa: E501
+    #   $ git remote show origin
+    cmd = ['git', 'remote', 'show', remote]
+    result = run(cmd)
+    if not result:
+        return None
+    result_lines = result.split('\n')
+    branch = None
+    for result_line in result_lines:
+        # There is a two-space indentation
+        match = re.match('  HEAD branch: (.*)', result_line)
+        if match is not None:
+            branch = match[1]
+            break
+    return branch
+
+
 def get_commits_data(
     base: str,
     head: str,
@@ -387,13 +435,10 @@ def extract_name_and_email(
     :param name_and_email: the name and email string
     :return: the extracted (name, email) tuple, or `None` if it failed
     """
-    email_match = re.search('<(.*)>', name_and_email)
-    if email_match is None:
+    match = re.search('(.*) <(.*)>', name_and_email)
+    if match is None:
         return None
-    name_match = re.search('(.*) <', name_and_email)
-    if name_match is None:
-        return None
-    return name_match.group(1), email_match.group(1)
+    return match.group(1), match.group(2)
 
 
 def get_env_var(
@@ -443,15 +488,6 @@ class CommitInfo:
         self.author_email = author_email
         self.is_merge_commit = is_merge_commit
 
-    def __repr__(self) -> str:  # noqa: D105
-        s = (
-            f'hash: {self.hash}\ntitle: {self.title}\nbody: {self.body}\n'
-            f'author: {self.author_name} <{self.author_email}>'
-        )
-        if self.is_merge_commit:
-            s += '\n(merge commit)'
-        return s
-
 
 class CommitDataRetriever:
     """
@@ -464,11 +500,11 @@ class CommitDataRetriever:
 
     def name(self) -> str:
         """Get a name that represents this retriever."""
-        raise NotImplementedError
+        raise NotImplementedError  # pragma: no cover
 
     def applies(self) -> bool:
         """Check if this retriever applies, i.e. can provide commit data."""
-        raise NotImplementedError
+        raise NotImplementedError  # pragma: no cover
 
     def get_commit_range(self) -> Optional[Tuple[str, str]]:
         """
@@ -478,34 +514,27 @@ class CommitDataRetriever:
 
         :return the (last commit that was checked, latest commit) tuple, or `None` if it failed
         """
-        raise NotImplementedError
+        raise NotImplementedError  # pragma: no cover
 
     def get_commits(self, base: str, head: str, **kwargs) -> Optional[List[CommitInfo]]:
         """Get commit data."""
-        raise NotImplementedError
+        raise NotImplementedError  # pragma: no cover
 
 
 class GitRetriever(CommitDataRetriever):
     """Implementation for any git repository."""
 
-    def name(self) -> str:
-        """Get a name that represents this retriever."""
+    def name(self) -> str:  # noqa: D102
         return 'git (default)'
 
-    def applies(self) -> bool:
-        """Check if this retriever applies, i.e. can provide commit data."""
+    def applies(self) -> bool:  # noqa: D102
         # Unless we only have access to a partial commit history
         return True
 
-    def get_commit_range(self) -> Optional[Tuple[str, str]]:
-        """
-        Get the range of commits to be checked: (last commit that was checked, latest commit).
-
-        The range excludes the first commit, e.g. ]first commit, second commit]
-
-        :return the (last commit that was checked, latest commit) tuple, or `None` if it failed
-        """
-        commit_hash_base = get_common_ancestor_commit_hash(options.default_branch)
+    def get_commit_range(self) -> Optional[Tuple[str, str]]:  # noqa: D102
+        default_branch = options.default_branch
+        logger.verbose_print(f"\tusing default branch '{default_branch}'")
+        commit_hash_base = get_common_ancestor_commit_hash(default_branch)
         if commit_hash_base is None:
             return None
         commit_hash_head = get_head_commit_hash()
@@ -519,8 +548,7 @@ class GitRetriever(CommitDataRetriever):
         head: str,
         check_merge_commits: bool = False,
         **kwargs,
-    ) -> Optional[List[CommitInfo]]:
-        """Get commit data."""
+    ) -> Optional[List[CommitInfo]]:  # noqa: D102
         ignore_merge_commits = not check_merge_commits
         commits_data = get_commits_data(base, head, ignore_merge_commits=ignore_merge_commits)
         individual_commits = split_commits_data(commits_data)
@@ -544,31 +572,22 @@ class GitRetriever(CommitDataRetriever):
                     commit_body,
                     author_name,
                     author_email,
-                    is_merge_commit
+                    is_merge_commit,
                 )
             )
         return commits
 
 
-class GitlabRetriever(CommitDataRetriever):
+class GitlabRetriever(GitRetriever):
     """Implementation for GitLab CI."""
 
-    def name(self) -> str:
-        """Get a name that represents this retriever."""
+    def name(self) -> str:  # noqa: D102
         return 'GitLab'
 
-    def applies(self) -> bool:
-        """Check if this retriever applies, i.e. can provide commit data."""
+    def applies(self) -> bool:  # noqa: D102
         return get_env_var('GITLAB_CI', print_if_not_found=False) is not None
 
-    def get_commit_range(self) -> Optional[Tuple[str, str]]:
-        """
-        Get the range of commits to be checked: (last commit that was checked, latest commit).
-
-        The range excludes the first commit, e.g. ]first commit, second commit]
-
-        :return the (last commit that was checked, latest commit) tuple, or `None` if it failed
-        """
+    def get_commit_range(self) -> Optional[Tuple[str, str]]:  # noqa: D102
         # See: https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
         default_branch = get_env_var('CI_DEFAULT_BRANCH', default=options.default_branch)
 
@@ -587,6 +606,32 @@ class GitlabRetriever(CommitDataRetriever):
             if commit_hash_base is None:
                 return None
             return commit_hash_base, commit_hash_head
+        elif get_env_var('CI_MERGE_REQUEST_ID', print_if_not_found=False):
+            # Get merge request target branch
+            target_branch = get_env_var('CI_MERGE_REQUEST_TARGET_BRANCH_NAME')
+            if target_branch is None:
+                return None
+            logger.verbose_print(
+                f"\ton merge request branch '{current_branch}': "
+                f"will check new commits off of target branch '{target_branch}'"
+            )
+            target_branch_sha = get_env_var('CI_MERGE_REQUEST_TARGET_BRANCH_SHA')
+            if target_branch_sha is None:
+                return None
+            return target_branch_sha, commit_hash_head
+        elif get_env_var('CI_EXTERNAL_PULL_REQUEST_IID', print_if_not_found=False):
+            # Get external merge request target branch
+            target_branch = get_env_var('CI_EXTERNAL_PULL_REQUEST_TARGET_BRANCH_NAME')
+            if target_branch is None:
+                return None
+            logger.verbose_print(
+                f"\ton merge request branch '{current_branch}': "
+                f"will check new commits off of target branch '{target_branch}'"
+            )
+            target_branch_sha = get_env_var('CI_EXTERNAL_PULL_REQUEST_TARGET_BRANCH_SHA')
+            if target_branch_sha is None:
+                return None
+            return target_branch_sha, commit_hash_head
         else:
             # Otherwise test all commits off of the default branch
             logger.verbose_print(
@@ -605,156 +650,43 @@ class GitlabRetriever(CommitDataRetriever):
                 return None
             return commit_hash_base, commit_hash_head
 
-    def get_commits(self, base: str, head: str, **kwargs) -> Optional[List[CommitInfo]]:
-        """Get commit data."""
-        return GitRetriever().get_commits(base, head, **kwargs)
 
-
-class CircleRetriever(CommitDataRetriever):
+class CircleCiRetriever(GitRetriever):
     """Implementation for CircleCI."""
 
-    def name(self) -> str:
-        """Get a name that represents this retriever."""
+    def name(self) -> str:  # noqa: D102
         return 'CircleCI'
 
-    def applies(self) -> bool:
-        """Check if this retriever applies, i.e. can provide commit data."""
+    def applies(self) -> bool:  # noqa: D102
         return get_env_var('CIRCLECI', print_if_not_found=False) is not None
 
-    def get_commit_range(self) -> Optional[Tuple[str, str]]:
-        """
-        Get the range of commits to be checked: (last commit that was checked, latest commit).
-
-        The range excludes the first commit, e.g. ]first commit, second commit]
-
-        :return the (last commit that was checked, latest commit) tuple, or `None` if it failed
-        """
+    def get_commit_range(self) -> Optional[Tuple[str, str]]:  # noqa: D102
         # See: https://circleci.com/docs/2.0/env-vars/#built-in-environment-variables
-        # TODO replace
         default_branch = options.default_branch
 
         commit_hash_head = get_env_var('CIRCLE_SHA1')
         if commit_hash_head is None:
             return None
 
-        # TODO support testing only new commits on the default branch
-        current_branch = get_env_var('CIRCLE_BRANCH')
-
-        # Test all commits off of the default branch
-        logger.verbose_print(
-            f"\ton branch '{current_branch}': "
-            f"will check forked commits off of default branch '{default_branch}'"
-        )
-        # Fetch default branch
-        remote = options.default_remote
-        if 0 != fetch_branch(default_branch, remote):
-            logger.print(f"failed to fetch '{default_branch}' from remote '{remote}'")
-            return None
-        # Use remote default branch ref
-        remote_branch_ref = remote + '/' + default_branch
-        commit_hash_base = get_common_ancestor_commit_hash(remote_branch_ref)
-        if commit_hash_base is None:
-            return None
-        return commit_hash_base, commit_hash_head
-
-    def get_commits(self, base: str, head: str, **kwargs) -> Optional[List[CommitInfo]]:
-        """Get commit data."""
-        return GitRetriever().get_commits(base, head, **kwargs)
-
-
-class AzurePipelinesRetriever(CommitDataRetriever):
-    """Implementation for Azure Pipelines."""
-
-    def name(self) -> str:
-        """Get a name that represents this retriever."""
-        return 'Azure Pipelines'
-
-    def applies(self) -> bool:
-        """Check if this retriever applies, i.e. can provide commit data."""
-        return get_env_var('TF_BUILD', print_if_not_found=False) is not None
-
-    def get_commit_range(self) -> Optional[Tuple[str, str]]:
-        """
-        Get the range of commits to be checked: (last commit that was checked, latest commit).
-
-        The range excludes the first commit, e.g. ]first commit, second commit]
-
-        :return the (last commit that was checked, latest commit) tuple, or `None` if it failed
-        """
-        # See: https://docs.microsoft.com/en-us/azure/devops/pipelines/build/variables?view=azure-devops&tabs=yaml#build-variables  # noqa: E501
-        # TODO replace
-        default_branch = options.default_branch
-
-        commit_hash_head = get_env_var('BUILD_SOURCEVERSION')
-        if commit_hash_head is None:
-            return None
-
-        # TODO support testing only new commits on the default branch
-        current_branch = get_env_var('BUILD_SOURCEBRANCHNAME')
-
-        # TODO use 'System.PullRequest.TargetBranch' and 'System.PullRequest.PullRequestId'
-
-        # Test all commits off of the default branch
-        logger.verbose_print(
-            f"\ton branch '{current_branch}': "
-            f"will check forked commits off of default branch '{default_branch}'"
-        )
-        # Fetch default branch
-        remote = options.default_remote
-        if 0 != fetch_branch(default_branch, remote):
-            logger.print(f"failed to fetch '{default_branch}' from remote '{remote}'")
-            return None
-        # Use remote default branch ref
-        remote_branch_ref = remote + '/' + default_branch
-        commit_hash_base = get_common_ancestor_commit_hash(remote_branch_ref)
-        if commit_hash_base is None:
-            return None
-        return commit_hash_base, commit_hash_head
-
-    def get_commits(self, base: str, head: str, **kwargs) -> Optional[List[CommitInfo]]:
-        """Get commit data."""
-        return GitRetriever().get_commits(base, head, **kwargs)
-
-
-class AppVeyorRetriever(CommitDataRetriever):
-    """Implementation for AppVeyor."""
-
-    def name(self) -> str:
-        """Get a name that represents this retriever."""
-        return 'AppVeyor'
-
-    def applies(self) -> bool:
-        """Check if this retriever applies, i.e. can provide commit data."""
-        return get_env_var('APPVEYOR', print_if_not_found=False) is not None
-
-    def get_commit_range(self) -> Optional[Tuple[str, str]]:
-        """
-        Get the range of commits to be checked: (last commit that was checked, latest commit).
-
-        The range excludes the first commit, e.g. ]first commit, second commit]
-
-        :return the (last commit that was checked, latest commit) tuple, or `None` if it failed
-        """
-        # See: https://www.appveyor.com/docs/environment-variables/
-        # TODO review this
-        default_branch = options.default_branch
-
-        commit_hash_head = get_env_var('APPVEYOR_PULL_REQUEST_HEAD_COMMIT')
-        if commit_hash_head is None:
-            commit_hash_head = get_head_commit_hash()
-
-        # If we're on the default branch, just test new commits
-        current_branch = get_env_var('APPVEYOR_PULL_REQUEST_HEAD_REPO_BRANCH')
-        if current_branch is not None and current_branch == default_branch:
+        # Check if base revision is provided to the environment, e.g.
+        #   environment:
+        #     CIRCLE_BASE_REVISION: << pipeline.git.base_revision >>
+        # See:
+        #   https://circleci.com/docs/2.0/pipeline-variables/
+        #   https://circleci.com/docs/2.0/env-vars/#built-in-environment-variables
+        base_revision = get_env_var('CIRCLE_BASE_REVISION', print_if_not_found=False)
+        if base_revision:
+            # For PRs, this is the commit of the base branch,
+            # and, for pushes to a branch, this is the commit before the new commits
             logger.verbose_print(
-                f"\ton default branch '{current_branch}': will check new commits"
+                f"\tchecking commits off of base revision '{base_revision}'"
             )
-            commit_hash_base = get_env_var('CI_COMMIT_BEFORE_SHA')
-            if commit_hash_base is None:
-                return None
-            return commit_hash_base, commit_hash_head
+            return base_revision, commit_hash_head
         else:
-            # Otherwise test all commits off of the default branch
+            current_branch = get_env_var('CIRCLE_BRANCH')
+            if current_branch is None:
+                return None
+            # Test all commits off of the default branch
             logger.verbose_print(
                 f"\ton branch '{current_branch}': "
                 f"will check forked commits off of default branch '{default_branch}'"
@@ -771,30 +703,126 @@ class AppVeyorRetriever(CommitDataRetriever):
                 return None
             return commit_hash_base, commit_hash_head
 
-    def get_commits(self, base: str, head: str, **kwargs) -> Optional[List[CommitInfo]]:
-        """Get commit data."""
-        return GitRetriever().get_commits(base, head, **kwargs)
+
+class AzurePipelinesRetriever(GitRetriever):
+    """Implementation for Azure Pipelines."""
+
+    def name(self) -> str:  # noqa: D102
+        return 'Azure Pipelines'
+
+    def applies(self) -> bool:  # noqa: D102
+        return get_env_var('TF_BUILD', print_if_not_found=False) is not None
+
+    def get_commit_range(self) -> Optional[Tuple[str, str]]:  # noqa: D102
+        # See: https://docs.microsoft.com/en-us/azure/devops/pipelines/build/variables?view=azure-devops&tabs=yaml#build-variables  # noqa: E501
+        commit_hash_head = get_env_var('BUILD_SOURCEVERSION')
+        if commit_hash_head is None:
+            return None
+        current_branch = get_env_var('BUILD_SOURCEBRANCHNAME')
+        if current_branch is None:
+            return None
+
+        base_branch = None
+        # Check if pull request
+        is_pull_request = get_env_var(
+            'SYSTEM_PULLREQUEST_PULLREQUESTID',
+            print_if_not_found=False,
+        )
+        if is_pull_request:
+            # Test all commits off of the target branch
+            target_branch = get_env_var('SYSTEM_PULLREQUEST_TARGETBRANCH')
+            if target_branch is None:
+                return None
+            logger.verbose_print(
+                f"\ton pull request branch '{current_branch}': "
+                f"will check forked commits off of target branch '{target_branch}'"
+            )
+            base_branch = target_branch
+        else:
+            # Test all commits off of the default branch
+            default_branch = options.default_branch
+            logger.verbose_print(
+                f"\ton branch '{current_branch}': "
+                f"will check forked commits off of default branch '{default_branch}'"
+            )
+            base_branch = default_branch
+        # Fetch base branch
+        assert base_branch
+        remote = options.default_remote
+        if 0 != fetch_branch(base_branch, remote):
+            logger.print(f"failed to fetch '{base_branch}' from remote '{remote}'")
+            return None
+        # Use remote default branch ref
+        remote_branch_ref = remote + '/' + base_branch
+        commit_hash_base = get_common_ancestor_commit_hash(remote_branch_ref)
+        if commit_hash_base is None:
+            return None
+        return commit_hash_base, commit_hash_head
+
+
+class AppVeyorRetriever(GitRetriever):
+    """Implementation for AppVeyor."""
+
+    def name(self) -> str:  # noqa: D102
+        return 'AppVeyor'
+
+    def applies(self) -> bool:  # noqa: D102
+        return get_env_var('APPVEYOR', print_if_not_found=False) is not None
+
+    def get_commit_range(self) -> Optional[Tuple[str, str]]:  # noqa: D102
+        # See: https://www.appveyor.com/docs/environment-variables/
+        default_branch = options.default_branch
+
+        commit_hash_head = get_env_var('APPVEYOR_REPO_COMMIT')
+        if commit_hash_head is None:
+            commit_hash_head = get_head_commit_hash()
+            if commit_hash_head is None:
+                return None
+
+        branch = get_env_var('APPVEYOR_REPO_BRANCH')
+        if branch is None:
+            return None
+
+        # Check if pull request
+        if get_env_var('APPVEYOR_PULL_REQUEST_NUMBER', print_if_not_found=False):
+            current_branch = get_env_var('APPVEYOR_PULL_REQUEST_HEAD_REPO_BRANCH')
+            if current_branch is None:
+                return None
+            target_branch = branch
+            logger.verbose_print(
+                f"\ton pull request branch '{current_branch}': "
+                f"will check commits off of target branch '{target_branch}'"
+            )
+            commit_hash_head = get_env_var('APPVEYOR_PULL_REQUEST_HEAD_COMMIT') or commit_hash_head
+            if commit_hash_head is None:
+                return None
+            commit_hash_base = get_common_ancestor_commit_hash(target_branch)
+            if commit_hash_base is None:
+                return None
+            return commit_hash_base, commit_hash_head
+        else:
+            # Otherwise test all commits off of the default branch
+            current_branch = branch
+            logger.verbose_print(
+                f"\ton branch '{current_branch}': "
+                f"will check forked commits off of default branch '{default_branch}'"
+            )
+            commit_hash_base = get_common_ancestor_commit_hash(default_branch)
+            if commit_hash_base is None:
+                return None
+            return commit_hash_base, commit_hash_head
 
 
 class GitHubRetriever(CommitDataRetriever):
     """Implementation for GitHub CI."""
 
-    def name(self) -> str:
-        """Get a name that represents this retriever."""
+    def name(self) -> str:  # noqa: D102
         return 'GitHub CI'
 
-    def applies(self) -> bool:
-        """Check if this retriever applies, i.e. can provide commit data."""
+    def applies(self) -> bool:  # noqa: D102
         return get_env_var('GITHUB_ACTIONS', print_if_not_found=False) == 'true'
 
-    def get_commit_range(self) -> Optional[Tuple[str, str]]:
-        """
-        Get the range of commits to be checked: (last commit that was checked, latest commit).
-
-        The range excludes the first commit, e.g. ]first commit, second commit]
-
-        :return the (last commit that was checked, latest commit) tuple, or `None` if it failed
-        """
+    def get_commit_range(self) -> Optional[Tuple[str, str]]:  # noqa: D102
         # See: https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
         self.github_token = get_env_var('GITHUB_TOKEN')
         if self.github_token is None:
@@ -840,13 +868,17 @@ class GitHubRetriever(CommitDataRetriever):
             else:
                 commit_hash_base = self.event_payload['before']
             commit_hash_head = self.event_payload['head_commit']['id']
-        else:
+        else:  # pragma: no cover
             logger.print('Unknown workflow event:', event_name)
             return None
         return commit_hash_base, commit_hash_head
 
-    def get_commits(self, base: str, head: str, **kwargs) -> Optional[List[CommitInfo]]:
-        """Get commit data."""
+    def get_commits(
+        self,
+        base: str,
+        head: str,
+        **kwargs,
+    ) -> Optional[List[CommitInfo]]:  # noqa: D102
         # Request commit data
         compare_url_template = self.event_payload['repository']['compare_url']
         compare_url = compare_url_template.format(base=base, head=head)
@@ -857,7 +889,7 @@ class GitHubRetriever(CommitDataRetriever):
         }
         connection.request('GET', compare_url, headers=headers)
         response = connection.getresponse()
-        if 200 != response.getcode():
+        if 200 != response.getcode():  # pragma: no cover
             from pprint import pprint
             logger.print('Request failed: compare_url')
             logger.print('reponse:', pprint(response.read().decode()))
@@ -911,7 +943,7 @@ def process_commits(
         logger.verbose_print(
             '\t' + commit.hash + (' (merge commit)' if commit.is_merge_commit else '')
         )
-        logger.verbose_print('\t' + commit.author_name, commit.author_email)
+        logger.verbose_print('\t' + (commit.author_name or 'N/A'), (commit.author_email or 'N/A'))
         logger.verbose_print('\t' + commit.title)
         logger.verbose_print('\t' + '\n\t'.join(commit.body))
 
@@ -978,17 +1010,24 @@ def check_infractions(
     return 0
 
 
-def main() -> int:
+def main(argv: Optional[List[str]] = None) -> int:
     """
     Entrypoint.
 
+    :param argv: the arguments to use, or `None` for sys.argv
     :return: 0 if successful, non-zero otherwise
     """
-    args = parse_args()
+    args = parse_args(argv)
     options.set_options(args)
-    options.apply_env_vars()
     logger.set_options(options)
 
+    # Print options
+    if options.verbose:
+        logger.verbose_print('Options:')
+        for name, value in options.get_options().items():
+            logger.verbose_print(f'\t{name}: {str(value)}')
+        logger.verbose_print()
+
     # Detect CI
     # Use first one that applies
     retrievers = [
@@ -996,7 +1035,7 @@ def main() -> int:
         GitHubRetriever,
         AzurePipelinesRetriever,
         AppVeyorRetriever,
-        CircleRetriever,
+        CircleCiRetriever,
         GitRetriever,
     ]
     commit_retriever = None
@@ -1007,6 +1046,15 @@ def main() -> int:
             break
     logger.print('Detected:', commit_retriever.name())
 
+    # Get default branch from remote if enabled
+    if options.default_branch_from_remote:
+        remote_default_branch = get_default_branch_from_remote(options.default_remote)
+        if remote_default_branch is None:
+            logger.print('Could not get default branch from remote')
+            return 1
+        options.default_branch = remote_default_branch
+        logger.print(f"\tgot default branch '{remote_default_branch}' from remote")
+
     # Get range of commits
     commit_range = commit_retriever.get_commit_range()
     if commit_range is None:
@@ -1043,5 +1091,5 @@ def main() -> int:
     return result
 
 
-if __name__ == '__main__':
+if __name__ == '__main__':  # pragma: no cover
     sys.exit(main())
